/*
 * Copyright (c) 2008 appspy.org, Contributors.
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0 
 * Unless required by applicable law or agreed to in writing, software distributed under the 
 * License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, 
 * either express or implied. See the License for the specific language governing permissions 
 * and limitations under the License. */

package org.appspy.server.dao.impl;

import java.util.Collection;

import javax.persistence.EntityManager;
import javax.persistence.PersistenceException;
import javax.persistence.Query;

import org.appspy.server.bo.Application;
import org.appspy.server.bo.ApplicationVersion;
import org.appspy.server.bo.ClientHost;
import org.appspy.server.bo.DelayType;
import org.appspy.server.bo.Environment;
import org.appspy.server.bo.HitResult;
import org.appspy.server.bo.HitURL;
import org.appspy.server.bo.Session;
import org.appspy.server.bo.URL;
import org.appspy.server.bo.User;
import org.appspy.server.bo.Webapp;
import org.appspy.server.dao.PerfDao;
import org.appspy.server.dao.xml.XmlDataLoader;
import org.springframework.beans.BeansException;
import org.springframework.beans.factory.BeanFactory;
import org.springframework.beans.factory.BeanFactoryAware;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.orm.jpa.JpaCallback;
import org.springframework.orm.jpa.support.JpaDaoSupport;
import org.springframework.transaction.annotation.Propagation;
import org.springframework.transaction.annotation.Transactional;

import com.thoughtworks.xstream.XStream;

/**
 * @author Olivier HEDIN / olivier@appspy.org
 */
public class PerfDaoImpl extends JpaDaoSupport implements PerfDao, InitializingBean, BeanFactoryAware {

	protected BeanFactory mBeanFactory = null;

	/* (non-Javadoc)
	 * @see org.appspy.server.dao.PerfDao#findAllWebapps()
	 */
	@SuppressWarnings("unchecked")
	public Collection<Webapp> findAllWebapps() {
		return getJpaTemplate().find("from Webapp");
	}

	/* (non-Javadoc)
	 * @see org.appspy.server.dao.PerfDao#findOrCreateWebapp(java.lang.String, org.appspy.server.bo.Application)
	 */
	public Webapp findOrCreateWebapp(String webappName, Application app) {
		
		Webapp result = findWebappByName(webappName, app);
		
		if (result == null)
		{
			result = new Webapp();
			result.setName(webappName);
			result.setApplication(app);
			
			saveWebapp(result);
		}
		
		return result;
	}

	/* (non-Javadoc)
	 * @see org.appspy.server.dao.PerfDao#findWebappByName(java.lang.String, org.appspy.server.bo.Application)
	 */
	@SuppressWarnings("unchecked")
	public Webapp findWebappByName(final String webappName, final Application app) {
		
		Collection<Webapp> queryResult = getJpaTemplate().executeFind(new JpaCallback(){

			public Object doInJpa(EntityManager em) throws PersistenceException {
				Query query = em.createQuery("from Webapp as w where w.mApplication.mId = :appId and w.mName = :webappName");
				query.setParameter("appId", app.getId());
				query.setParameter("webappName", webappName);
				return query.getResultList();
			}});
		
		Webapp result = null;
		
		if (!queryResult.isEmpty())
		{
			result = queryResult.iterator().next();
		}
		
		return result;
	}

	/* (non-Javadoc)
	 * @see org.appspy.server.dao.PerfDao#saveWebapp(org.appspy.server.bo.Webapp)
	 */
	public Webapp saveWebapp(Webapp webapp) {
		getJpaTemplate().persist(webapp);
		
		return webapp;
	}

	/* (non-Javadoc)
	 * @see org.appspy.server.dao.PerfDao#updateWebapp(org.appspy.server.bo.Webapp)
	 */
	public Webapp updateWebapp(Webapp webapp) {
		return getJpaTemplate().merge(webapp);
	}

	/* (non-Javadoc)
	 * @see org.appspy.server.dao.PerfDao#findAllUsers()
	 */
	@SuppressWarnings("unchecked")
	public Collection<User> findAllUsers() {
		return getJpaTemplate().find("from User");
	}

	/* (non-Javadoc)
	 * @see org.appspy.server.dao.PerfDao#findUserByName(java.lang.String)
	 */
	@SuppressWarnings("unchecked")
	public User findUserByName(final String userName) {
		Collection<User> queryResult = getJpaTemplate().executeFind(new JpaCallback(){

			public Object doInJpa(EntityManager em) throws PersistenceException {
				Query query = em.createQuery("from User as u where u.mUserName = :userName");
				query.setParameter("userName", userName);
				return query.getResultList();
			}});
		
		User result = null;
		
		if (!queryResult.isEmpty())
		{
			result = queryResult.iterator().next();
		}
		
		return result;
	}

	/* (non-Javadoc)
	 * @see org.appspy.server.dao.PerfDao#saveUser(org.appspy.server.bo.User)
	 */
	public User saveUser(User user) {
		getJpaTemplate().persist(user);
		return user;
	}

	/* (non-Javadoc)
	 * @see org.appspy.server.dao.PerfDao#updateUser(org.appspy.server.bo.User)
	 */
	public User updateUser(User user) {
		return getJpaTemplate().merge(user);
	}

	/* (non-Javadoc)
	 * @see org.appspy.server.dao.PerfDao#findOrCreateUser(java.lang.String)
	 */
	public User findOrCreateUser(String userName) {
		User result = findUserByName(userName);
		
		if (result == null)
		{
			result = new User();
			result.setUserName(userName);
			saveUser(result);
		}
		
		return result;
	}

	/* (non-Javadoc)
	 * @see org.appspy.server.dao.PerfDao#findAllClientHosts()
	 */
	@SuppressWarnings("unchecked")
	public Collection<ClientHost> findAllClientHosts() {
		return getJpaTemplate().find("from ClientHost");
	}

	/* (non-Javadoc)
	 * @see org.appspy.server.dao.PerfDao#findClientHostByIP(java.lang.String)
	 */
	@SuppressWarnings("unchecked")
	public ClientHost findClientHostByIP(final String ip) {
		Collection<ClientHost> queryResult = getJpaTemplate().executeFind(new JpaCallback(){

			public Object doInJpa(EntityManager em) throws PersistenceException {
				Query query = em.createQuery("from ClientHost as c where c.mClientHostIP = :ip");
				query.setParameter("ip", ip);
				return query.getResultList();
			}});
		
		ClientHost result = null;
		
		if (!queryResult.isEmpty())
		{
			result = queryResult.iterator().next();
		}
		
		return result;
	}

	/* (non-Javadoc)
	 * @see org.appspy.server.dao.PerfDao#findOrCreateClientHost(java.lang.String)
	 */
	public ClientHost findOrCreateClientHost(String ip) {
		ClientHost result = findClientHostByIP(ip);
		
		if (result == null)
		{
			result = new ClientHost();
			result.setClientHostIP(ip);
			saveClientHost(result);
		}
		
		return result;
	}

	/* (non-Javadoc)
	 * @see org.appspy.server.dao.PerfDao#saveClientHost(org.appspy.server.bo.ClientHost)
	 */
	public ClientHost saveClientHost(ClientHost clientHost) {
		getJpaTemplate().persist(clientHost);
		return clientHost;
	}

	/* (non-Javadoc)
	 * @see org.appspy.server.dao.PerfDao#updateClientHost(org.appspy.server.bo.ClientHost)
	 */
	public ClientHost updateClientHost(ClientHost clientHost) {
		return getJpaTemplate().merge(clientHost);
	}

	/* (non-Javadoc)
	 * @see org.appspy.server.dao.PerfDao#findAllSessions()
	 */
	@SuppressWarnings("unchecked")
	public Collection<Session> findAllSessions() {
		return getJpaTemplate().find("from Session");
	}

	/* (non-Javadoc)
	 * @see org.appspy.server.dao.PerfDao#findClientHostById(java.lang.Long)
	 */
	public ClientHost findClientHostById(Long clientHostId) {
		return getJpaTemplate().find(ClientHost.class, clientHostId);
	}

	/* (non-Javadoc)
	 * @see org.appspy.server.dao.PerfDao#findSession(java.lang.String, org.appspy.server.bo.ApplicationVersion, org.appspy.server.bo.Environment, org.appspy.server.bo.User)
	 */
	@SuppressWarnings("unchecked")
	public Session findSession(final String sessionCode, final ApplicationVersion version,
			final Environment env, final User user) {
		Collection<Session> queryResult = getJpaTemplate().executeFind(new JpaCallback(){

			public Object doInJpa(EntityManager em) throws PersistenceException {
				Query query = em.createQuery("from Session as s where s.mCode = :sessionCode and s.mVersion.mId = :versionId and s.mEnvironment.mId = :envId and s.mUser.mId = :userId");
				query.setParameter("sessionCode", sessionCode);
				query.setParameter("versionId", version.getId());
				query.setParameter("envId", env.getId());
				query.setParameter("userId", user.getId());
				
				return query.getResultList();
			}});
		
		Session result = null;
		
		if (!queryResult.isEmpty())
		{
			result = queryResult.iterator().next();
		}
		
		return result;
	}
	
	/* (non-Javadoc)
	 * @see org.appspy.server.dao.PerfDao#findOrCreateSession(java.lang.String, org.appspy.server.bo.ApplicationVersion, org.appspy.server.bo.Environment, org.appspy.server.bo.User)
	 */
	public Session findOrCreateSession(String sessionCode,
			ApplicationVersion version, Environment env, User user, ClientHost clientHost) {
		Session result = findSession(sessionCode, version, env, user);
		
		if (result == null)
		{
			result = new Session();
			result.setCode(sessionCode);
			result.setClientHost(clientHost);
			result.setEnvironment(env);
			result.setUser(user);
			result.setVersion(version);
			saveSession(result);
		}
		else
		{
			if (result.getClientHost() == null)
			{
				result.setClientHost(clientHost);
				updateSession(result);
			}
		}
		
		return result;
	}

	/* (non-Javadoc)
	 * @see org.appspy.server.dao.PerfDao#findSessionById(java.lang.Long)
	 */
	public Session findSessionById(Long sessionId) {
		return getJpaTemplate().find(Session.class, sessionId);
	}

	/* (non-Javadoc)
	 * @see org.appspy.server.dao.PerfDao#findUserById(java.lang.Long)
	 */
	public User findUserById(Long userId) {
		return getJpaTemplate().find(User.class, userId);
	}

	/* (non-Javadoc)
	 * @see org.appspy.server.dao.PerfDao#findWebappById(java.lang.Long)
	 */
	public Webapp findWebappById(Long webappId) {
		return getJpaTemplate().find(Webapp.class, webappId);
	}

	/* (non-Javadoc)
	 * @see org.appspy.server.dao.PerfDao#saveSession(org.appspy.server.bo.Session)
	 */
	public Session saveSession(Session session) {
		getJpaTemplate().persist(session);
		return session;
	}

	/* (non-Javadoc)
	 * @see org.appspy.server.dao.PerfDao#updateSession(org.appspy.server.bo.Session)
	 */
	public Session updateSession(Session session) {
		return getJpaTemplate().merge(session);
	}

	/* (non-Javadoc)
	 * @see org.appspy.server.dao.PerfDao#findAllURLs()
	 */
	@SuppressWarnings("unchecked")
	public Collection<URL> findAllURLs() {
		return getJpaTemplate().find("from URL");
	}

	/* (non-Javadoc)
	 * @see org.appspy.server.dao.PerfDao#findOrCreateURL(java.lang.String, org.appspy.server.bo.Webapp)
	 */
	public URL findOrCreateURL(String urlName, Webapp webapp) {
		
		URL result = findURLByName(urlName, webapp);
		
		if (result == null)
		{
			result = new URL();
			result.setApplication(webapp.getApplication());
			result.setName(urlName);
			result.setWebapp(webapp);
			saveURL(result);
		}
		
		return result;
	}

	/* (non-Javadoc)
	 * @see org.appspy.server.dao.PerfDao#findURLById(java.lang.Long)
	 */
	public URL findURLById(Long urlId) {
		return getJpaTemplate().find(URL.class, urlId);
	}

	/* (non-Javadoc)
	 * @see org.appspy.server.dao.PerfDao#findURLByName(java.lang.String, org.appspy.server.bo.Webapp)
	 */
	@SuppressWarnings("unchecked")
	public URL findURLByName(final String urlName, final Webapp webapp) {
		Collection<URL> queryResult = getJpaTemplate().executeFind(new JpaCallback(){

			public Object doInJpa(EntityManager em) throws PersistenceException {
				Query query = em.createQuery("from URL as u where u.mName = :urlName and u.mWebapp = :webapp");
				query.setParameter("urlName", urlName);
				query.setParameter("webapp", webapp);
				
				return query.getResultList();
			}});
		
		URL result = null;
		
		if (!queryResult.isEmpty())
		{
			result = queryResult.iterator().next();
		}
		
		return result;
	}

	/* (non-Javadoc)
	 * @see org.appspy.server.dao.PerfDao#saveURL(org.appspy.server.bo.URL)
	 */
	public URL saveURL(URL url) {
		getJpaTemplate().persist(url);
		return url;
	}

	/* (non-Javadoc)
	 * @see org.appspy.server.dao.PerfDao#updateURL(org.appspy.server.bo.URL)
	 */
	public URL updateURL(URL url) {
		return getJpaTemplate().merge(url);
	}

	/* (non-Javadoc)
	 * @see org.appspy.server.dao.PerfDao#saveHitURL(org.appspy.server.bo.HitURL)
	 */
	public HitURL saveHitURL(HitURL hitURL) {
		getJpaTemplate().persist(hitURL);
		return hitURL;
	}
	
	
	
	/* (non-Javadoc)
     * @see org.appspy.server.dao.PerfDao#findAllDelayTypes()
     */
    @SuppressWarnings("unchecked")
    public Collection<DelayType> findAllDelayTypes()
    {
		return getJpaTemplate().find("from DelayType");
    }

	/* (non-Javadoc)
     * @see org.appspy.server.dao.PerfDao#findAllHitResults()
     */
    @SuppressWarnings("unchecked")
    public Collection<HitResult> findAllHitResults()
    {
		return getJpaTemplate().find("from HitResult");
    }

	/* (non-Javadoc)
     * @see org.appspy.server.dao.PerfDao#saveDelayType(org.appspy.server.bo.DelayType)
     */
	@Transactional(readOnly=false, propagation=Propagation.REQUIRED)
    public DelayType saveDelayType(DelayType delayType)
    {
		getJpaTemplate().persist(delayType);
		return delayType;
    }

	/* (non-Javadoc)
     * @see org.appspy.server.dao.PerfDao#saveHitResult(org.appspy.server.bo.HitResult)
     */
	@Transactional(readOnly=false, propagation=Propagation.REQUIRED)
    public HitResult saveHitResult(HitResult hitResult)
    {
		getJpaTemplate().persist(hitResult);
		return hitResult;
    }
    
    

	/* (non-Javadoc)
     * @see org.appspy.server.dao.PerfDao#findOrCreateDelayType(int)
     */
    public DelayType findOrCreateDelayType(int value)
    {
		DelayType result = findDelayTypeByValue(value);
		
		if (result == null)
		{
			result = new DelayType();
			result.setCode("Undefined");
			result.setValue(value);
			saveDelayType(result);
		}
		
		return result;
    }
    
	/* (non-Javadoc)
     * @see org.appspy.server.dao.PerfDao#findDelayTypeByValue(int)
     */
    @SuppressWarnings("unchecked")
    public DelayType findDelayTypeByValue(final int value)
    {
		Collection<DelayType> queryResult = getJpaTemplate().executeFind(new JpaCallback(){

			public Object doInJpa(EntityManager em) throws PersistenceException {
				Query query = em.createQuery("from DelayType as dt where dt.mValue = :value");
				query.setParameter("value", value);
				return query.getResultList();
			}});
		
		DelayType result = null;
		
		if (!queryResult.isEmpty())
		{
			result = queryResult.iterator().next();
		}
		
		return result;
    }

	/* (non-Javadoc)
     * @see org.appspy.server.dao.PerfDao#findOrCreateHitResult(int)
     */
    public HitResult findOrCreateHitResult(int value)
    {
		HitResult result = findHitResultByValue(value);
		
		if (result == null)
		{
			result = new HitResult();
			result.setCode("Undefined");
			result.setValue(value);
			saveHitResult(result);
		}
		
		return result;
    }

	/**
     * @param value
     * @return
     */
    @SuppressWarnings("unchecked")
    public HitResult findHitResultByValue(final int value)
    {
		Collection<HitResult> queryResult = getJpaTemplate().executeFind(new JpaCallback(){

			public Object doInJpa(EntityManager em) throws PersistenceException {
				Query query = em.createQuery("from HitResult as hr where hr.mValue = :value");
				query.setParameter("value", value);
				return query.getResultList();
			}});
		
		HitResult result = null;
		
		if (!queryResult.isEmpty())
		{
			result = queryResult.iterator().next();
		}
		
		return result;
    }

	/* (non-Javadoc)
	 * @see org.springframework.dao.support.DaoSupport#initDao()
	 */
	@SuppressWarnings("unchecked")
	@Override
	protected void initDao() throws Exception {
		super.initDao();
		
		Collection<HitResult> hitResults = findAllHitResults();
		
		if (hitResults == null || hitResults.size() == 0)
		{
			XStream xstream = new XStream();
			Collection<Object> initialDatas = (Collection<Object>) xstream.fromXML(getClass().getClassLoader().getResourceAsStream("org/appspy/server/xml/initial-perf-data.xml"));
			for (Object data : initialDatas) {
				XmlDataLoader dataLoader = (XmlDataLoader) mBeanFactory.getBean(data.getClass().getName() + ".dataLoader");
				dataLoader.loadData(data);
			}
		}
	}

	/* (non-Javadoc)
	 * @see org.springframework.beans.factory.BeanFactoryAware#setBeanFactory(org.springframework.beans.factory.BeanFactory)
	 */
	public void setBeanFactory(BeanFactory beanFactory) throws BeansException {
		mBeanFactory = beanFactory;
	}
}
